@@ -1,12 +1,14 @@
 import datetime
 import logging
-import os.path
 import pathlib
 import pickle
 import time
 import uuid
+import json
 import warnings
 
+from collections import Iterable
+
 import google.protobuf.timestamp_pb2
 import google.protobuf.wrappers_pb2
 
@@ -24,6 +26,13 @@ def build_timestamp(timestamp=None) -> google.protobuf.timestamp_pb2.Timestamp:
     return proto_timestamp.FromDatetime(timestamp or datetime.datetime.utcnow())
 
 
+def to_datetime(timestamp: google.protobuf.timestamp_pb2.Timestamp) -> datetime.datetime:
+    try:
+        return timestamp.ToDateTime()
+    except AttributeError:
+        pass
+
+
 class Run:
     """Protocol run (dummy)"""
     SERIALISATION_EXT = 'pkl'
@@ -35,12 +44,13 @@ class Run:
         self.protocol_id = protocol_id
         self.user_info = user_info
         self.args = list(args or ())
-        self.start_time = datetime.datetime.utcnow()
+        self._start_time = None
         self.end_time = None
         self.device = device
+        self._acquisition_run_ids = None
 
     @property
-    def serialisation_dir(self) -> str:
+    def serialisation_dir(self) -> pathlib.Path:
         return self.build_serialisation_dir(device=self.device)
 
     @property
@@ -49,32 +59,35 @@ class Run:
 
     @property
     def as_dict(self) -> dict:
+        """Convert to Python data types"""
+
         return dict(
             state=self.state,
             run_id=self.run_id,
             protocol_id=self.protocol_id,
+            _acquisition_run_ids=self._acquisition_run_ids,
             user_info=dict(
-                protocol_group_id=str(self.user_info.protocol_group_id),
-                sample_id=str(self.user_info.sample_id),
+                protocol_group_id=self.user_info.protocol_group_id.value,
+                sample_id=self.user_info.sample_id.value,
             ),
-            args=self.args,
-            start_time=self.start_time,
-            end_time=self.end_time,
+            args=list(self.args),
+            _start_time=self._start_time,
+            _end_time=self._end_time,
             device=self.device,
         )
 
     @property
-    def path(self) -> str:
-        return os.path.join(self.serialisation_dir, self.filename)
+    def path(self) -> pathlib.Path:
+        return self.serialisation_dir.joinpath(self.filename)
 
     @classmethod
-    def build_serialisation_dir(cls, device: dict) -> str:
-        return os.path.join(pyminknow.config.RUN_DIR, device['name'])
+    def build_serialisation_dir(cls, device: dict) -> pathlib.Path:
+        return pathlib.Path(pyminknow.config.RUN_DIR).joinpath(device['name'])
 
     def serialise(self):
-        os.makedirs(self.serialisation_dir, exist_ok=True)
+        self.serialisation_dir.mkdir(parents=True, exist_ok=True)
 
-        with open(self.path, 'wb') as file:
+        with self.path.open('wb') as file:
             pickle.dump(self.as_dict, file)
 
             LOGGER.info("Wrote '%s'", file.name)
@@ -86,7 +99,7 @@ class Run:
             setattr(self, attr, value)
 
     def load(self) -> dict:
-        with open(self.path, 'rb') as file:
+        with self.path.open('rb') as file:
             data = pickle.load(file)
 
             LOGGER.debug("Read '%s'", file.name)
@@ -96,19 +109,60 @@ class Run:
     def deserialise(self):
         self.from_dict(self.load())
 
+    def build_filenames(self) -> iter:
+        """Generate filenames for output data files"""
+        templates = {
+            'drift_correction_{flow_cell_id}_{acq}.csv',
+            'duty_time_{flow_cell_id}_{acq}.csv',
+            'final_summary_{flow_cell_id}_{acq}.txt',
+            'mux_scan_data_{flow_cell_id}_{acq}.csv',
+            'sequencing_summary_{flow_cell_id}_{acq}.txt',
+            'throughput_{flow_cell_id}_{acq}.csv',
+        }
+        for template in templates:
+            yield template.format(flow_cell_id=self.flow_cell_id, acq=self.acq_id_short)
+
+        templates = {
+            'report_{flow_cell_id}_{day}_{time}_{run_id_short}.md',
+            'report_{flow_cell_id}_{day}_{time}_{run_id_short}.pdf',
+        }
+        for template in templates:
+            yield template.format(flow_cell_id=self.flow_cell_id, acq=self.acq_id_short, day=self.day,
+                                  time=self.time, run_id_short=self.run_id_short)
+
     def save_data(self):
         """Write sequence data to disk"""
 
-        filename = 'my_data.txt'
-        path = os.path.join(self.output_directory, filename)
+        # Make subdirectories e.g. "fast5_fail"
+        for test in {'fast5', 'fastq'}:
+            for result in {'pass', 'fail'}:
+                subdir = self.output_path.joinpath("{}_{}".format(test, result))
 
-        # Write to disk
-        os.makedirs(self.output_directory, exist_ok=True)
-        with open(path, 'w') as file:
-            # Generate some dummy data
-            file.write('Hello world!\n')
+                # One folder per barcode containing an empty file
+                for i in range(25):
+                    barcode = "barcode" + str(i).zfill(2)
 
-            LOGGER.debug("Wrote '%s'", file.name)
+                    subsubdir = subdir.joinpath(barcode)
+
+                    subsubdir.mkdir(parents=True, exist_ok=True)
+
+                    filename = self.run_code + '_0.' + test
+                    subsubdir.joinpath(filename).touch()
+
+        # Create data files
+        for filename in self.build_filenames():
+            path = self.output_path.joinpath(filename)
+            path.touch()
+
+            LOGGER.debug("Wrote '%s'", path)
+
+    @property
+    def acq_id_short(self) -> str:
+        return self.acquisition_run_id.partition('-')[0]
+
+    @property
+    def run_id_short(self) -> str:
+        return self.run_id.partition('-')[0]
 
     @property
     def state(self) -> minknow.rpc.protocol_pb2.ProtocolState:
@@ -119,17 +173,22 @@ class Run:
         """Change state"""
         self._state = state
         LOGGER.debug('Run %s changed state to %s', self.run_id, self.state)
-        self.serialise()
 
     def start(self):
+        self.start_time = datetime.datetime.utcnow()
         self.state = minknow.rpc.protocol_pb2.ProtocolState.PROTOCOL_RUNNING
+        self.serialise()
+        self.run()
+        self.stop()
 
-        LOGGER.debug("Starting run ID: %s", self.run_id)
-        LOGGER.debug("Protocol group ID: %s", self.user_info.protocol_group_id.value)
-        LOGGER.debug("Sample ID: %s", self.user_info.sample_id.value)
+    def run(self):
+        LOGGER.debug("Starting run ID: '%s'", self.run_id)
+        time.sleep(pyminknow.config.RUN_DURATION)
 
+    def stop(self):
         self.save_data()
         self.finish()
+        self.serialise()
 
     def finish(self):
         self.end_time = datetime.datetime.utcnow()
@@ -140,32 +199,48 @@ class Run:
         return self.state == minknow.rpc.protocol_pb2.ProtocolState.PROTOCOL_COMPLETED
 
     @property
-    def output_directory(self) -> str:
+    def protocol_group_id(self) -> str:
+        return self.user_info.protocol_group_id.value
+
+    @property
+    def sample_id(self) -> str:
+        return self.user_info.sample_id.value
+
+    @property
+    def output_path(self) -> pathlib.Path:
         """
         The directory to save the sequencing data
-        e.g. "/data/external_run_id/external_run_id/protcol_run_id"
+        "/data/protocol_group_id/sample_id/<run code>"
+        e.g. "/data/covid19-20200512-1589296413/covid19-20200512-1589296413/20200512_1517_X2_FAN43172_fa3c47d5"
         """
-        external_run_id = str(self.user_info.protocol_group_id)
-        return os.path.join(
-            pyminknow.config.DATA_DIR,
-            external_run_id,
-            external_run_id,
-            self.run_id,
+        return pathlib.Path(pyminknow.config.DATA_DIR).joinpath(
+            self.protocol_group_id,
+            self.sample_id,
+            self.run_code,
         )
 
     @property
-    def data_subdir(self):
+    def flow_cell_id(self) -> str:
+        return self.device['flow_cell']['flow_cell_id']
+
+    @property
+    def run_code(self):
         """
-        Generate protocol run identifier
-        e.g. "DATE_TIME_DEVICE_FLOWCELLID_PARTOFAQUISITIONID"
+        Generate protocol run identifier "DATE_TIME_DEVICE_FLOWCELLID_PARTOFAQUISITIONID"
+        e.g. "20200512_1517_X2_FAN43172_fa3c47d5"
         """
         device_id = self.device['name']
-        flow_cell_id = self.device['flow_cell']['flow_cell_id']
-        day = self.start_time.date().strftime('%Y%m%d')
-        clock_time = self.start_time.strftime('%H%M')
-        unique = self.run_id.partition('-')[0]
-        return "{day}_{time}_{device}_{flow_cell}_{acq}".format(day=day, time=clock_time, device=device_id,
-                                                                flow_cell=flow_cell_id, acq=unique)
+        return "{day}_{time}_{device}_{flow_cell}_{aqu_short}".format(day=self.day, time=self.time, device=device_id,
+                                                                      flow_cell=self.flow_cell_id,
+                                                                      aqu_short=self.acq_id_short)
+
+    @property
+    def day(self) -> str:
+        return self._start_time.date().strftime('%Y%m%d')
+
+    @property
+    def time(self):
+        return self._start_time.strftime('%H%M')
 
     @classmethod
     def make_run_id(cls) -> str:
@@ -179,11 +254,11 @@ class Run:
         )
 
     @property
-    def start_time(self):
+    def start_time(self) -> google.protobuf.timestamp_pb2.Timestamp:
         return build_timestamp(self._start_time)
 
     @start_time.setter
-    def start_time(self, value):
+    def start_time(self, value: datetime.datetime):
         self._start_time = value
 
     @property
@@ -191,15 +266,16 @@ class Run:
         return build_timestamp(self._end_time)
 
     @end_time.setter
-    def end_time(self, value):
+    def end_time(self, value) -> google.protobuf.timestamp_pb2.Timestamp:
         self._end_time = value
 
     @property
     def info(self) -> minknow.rpc.protocol_pb2.ProtocolRunInfo:
         return minknow.rpc.protocol_pb2.ProtocolRunInfo(
+            run_id=self.run_id,
             protocol_id=self.protocol_id,
             args=self.args,
-            output_path=self.output_directory,
+            output_path=str(self.output_path),
             state=self.state,
             start_time=self.start_time,
             end_time=self.end_time,
@@ -207,9 +283,23 @@ class Run:
             acquisition_run_ids=self.acquisition_run_ids,
         )
 
+    @staticmethod
+    def build_acquisition_run_ids(n: int = 3):
+        return [str(uuid.uuid4()) for _ in range(n)]
+
     @property
     def acquisition_run_ids(self) -> list:
-        return [str(uuid.uuid4()) for _ in range(3)]
+        if not self._acquisition_run_ids:
+            self._acquisition_run_ids = self.build_acquisition_run_ids()
+        return self._acquisition_run_ids
+
+    @acquisition_run_ids.setter
+    def acquisition_run_ids(self, value):
+        self._acquisition_run_ids = value
+
+    @property
+    def acquisition_run_id(self) -> str:
+        return self.acquisition_run_ids[-1]
 
     @classmethod
     def get_run_ids(cls, device: dict):
@@ -217,7 +307,7 @@ class Run:
         directory = cls.build_serialisation_dir(device=device)
         paths = pathlib.Path(directory).glob('*.{}'.format(cls.SERIALISATION_EXT))
         yield from (
-            # remove file ext
+            # Remove file extension
             _path.stem for _path in
             # Sort by creation time
             sorted(paths, key=lambda _path: _path.stat().st_ctime)
@@ -252,22 +342,48 @@ class ProtocolService(minknow.rpc.protocol_pb2_grpc.ProtocolServiceServicer):
     def clear_protocol_cache(self):
         raise NotImplementedError
 
+    @staticmethod
+    def tag_value(value) -> minknow.rpc.protocol_pb2.ProtocolInfo.TagValue:
+        """
+        Translate a value in a Python native data type to a ProtocolInfo.TagValue (protocol buffers data type)
+
+        https://github.com/nanoporetech/minknow_lims_interface/blob/master/minknow/rpc/protocol.proto#L174
+        """
+
+        type_map = dict(
+            bool_value=bool,
+            string_value=str,
+            float_value=float,
+            int_value=int,
+            array_value=Iterable,
+        )
+
+        for arg, data_type in type_map.items():
+            if isinstance(value, data_type):
+
+                # Serialise arrays to JSON format
+                if arg == 'array_value':
+                    value = json.dumps(list(value))
+
+                kwargs = {arg: value}
+
+                return minknow.rpc.protocol_pb2.ProtocolInfo.TagValue(**kwargs)
+
+        raise ValueError(value)
+
     @staticmethod
     def get_protocol_info() -> list:
         """Build collection of ProtocolInfo objects"""
         return [
             minknow.rpc.protocol_pb2.ProtocolInfo(
-                identifier='sequencing/sequencing_MIN106_MIN107_RNA:FLO-MIN107:SQK-RNA002:True',
-                name=protocol_name,
-                tags={
-                    'flow cell': minknow.rpc.protocol_pb2.ProtocolInfo.TagValue(string_value="FLO-MIN106"),
-                    'kit': minknow.rpc.protocol_pb2.ProtocolInfo.TagValue(string_value="SQK-LSK109"),
-                    'experiment type': minknow.rpc.protocol_pb2.ProtocolInfo.TagValue(string_value="sequencing"),
-                    'kit_category': minknow.rpc.protocol_pb2.ProtocolInfo.TagValue(
-                        string_value='[\"RNA\",\"PCR-Free\",\"No Multiplexing\"]'),
-                    'barcoding': minknow.rpc.protocol_pb2.ProtocolInfo.TagValue(bool_value=False),
-                }
-            ) for protocol_name in pyminknow.config.PROTOCOLS
+                identifier=protocol['identifier'],
+                name=protocol['name'],
+                tags={key: ProtocolService.tag_value(value) for key, value in protocol['tags'].items()},
+                tag_extraction_result=minknow.rpc.protocol_pb2.ProtocolInfo.TagExtractionResult(
+                    success=True,
+                    error_report='',
+                ),
+            ) for protocol in pyminknow.config.PROTOCOLS
         ]
 
     def _start_protocol(self, identifier, user_info, args):
@@ -285,6 +401,18 @@ class ProtocolService(minknow.rpc.protocol_pb2_grpc.ProtocolServiceServicer):
 
         return minknow.rpc.protocol_pb2.StartProtocolResponse(run_id=run_id)
 
+    def stop_protocol(self, request, context) -> minknow.rpc.protocol_pb2.StopProtocolResponse:
+        """
+        Stops the currently running protocol script instance.
+
+        https://github.com/nanoporetech/minknow_lims_interface/blob/master/minknow/rpc/protocol.proto#L17
+        """
+
+        run = Run(Run.latest_run_id)
+        run.stop()
+
+        return minknow.rpc.protocol_pb2.StopProtocolResponse()
+
     @property
     def latest_run_id(self):
         """The identifier of the most recently-started run"""
@@ -294,14 +422,11 @@ class ProtocolService(minknow.rpc.protocol_pb2_grpc.ProtocolServiceServicer):
     def run_ids(self) -> list:
         return list(Run.get_run_ids(device=self.device))
 
-    def get_run_info(self, request, context):
+    def get_run_info(self, request, context) -> minknow.rpc.protocol_pb2.ProtocolRunInfo:
         run_id = request.run_id
 
         # If no run ID is provided, information about the most recently started protocol run is provided
-        if not run_id:
-            run_id = self.latest_run_id
-
-        run = Run(run_id=run_id, device=self.device)
+        run = Run(run_id=run_id or self.latest_run_id, device=self.device)
         run.deserialise()
 
         return run.info
